--------------------------------------------------------------------------------
Library IEEE;
Use IEEE.Std_Logic_1164.all;
Use IEEE.Std_Logic_Unsigned.All;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
Entity WB_LCDCTRL Is Port
   (
     -- Wishbone Interface for LCD controller
     STB_I           : In    Std_Logic;
     CYC_I           : In    Std_Logic;
     ACK_O           : Out   Std_Logic;
     ADR_I           : In    Std_Logic;
     WE_I            : In    Std_Logic;
     DAT_I           : In    Std_Logic_Vector(7 DownTo 0);
     DAT_O           : Out   Std_Logic_Vector(7 DownTo 0);
     CLK_I           : In    Std_Logic;
     RST_I           : In    Std_Logic;

     -- Interface to CTRL/LCD
     LCD_E           : Out   Std_Logic;
     LCD_DATA        : InOut Std_Logic_Vector(7 DownTo 0);
     LCD_RW          : Out   Std_Logic;
     LCD_RS          : Out   Std_Logic
   );
End WB_LCDCTRL;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
Architecture RTL Of WB_LCDCTRL Is
    Type TState Is
        (
           State_AddressSetup,
           State_SetupDelay1,
           State_SetupDelay2,
           State_Enable_LCD,
           State_ACK,
           State_PostACKDelay
        );

    Signal   State                : TState;
    Signal   State_Next           : TState;
    Signal   LCD_Read             : Std_Logic;
    Signal   LCD_Write            : Std_Logic;
    Signal   LCD_Select           : Std_Logic;
    Signal   CycleCounter         : Std_Logic_Vector(6 DownTo 0);
    Signal   CycleCounter_Reset   : Std_Logic;

Begin
    ----------------------------------------------------------------------------
    -- Permanent Assignments
    -- Driven outputs
    LCD_Select <= STB_I And CYC_I;
    LCD_Read   <= STB_I And CYC_I And (Not WE_I);
    LCD_Write  <= STB_I And CYC_I And (    WE_I);
    DAT_O      <= LCD_DATA;
    ----------------------------------------------------------------------------

    ----------------------------------------------------------------------------
    Multiplexers:
    Process (LCD_Write, DAT_I)
    Begin
        If LCD_Write = '1' Then
           LCD_DATA <= DAT_I;
        Else
           LCD_DATA <= (Others => 'Z');
        End If;
    End Process;
    ----------------------------------------------------------------------------

    ----------------------------------------------------------------------------
    FSM_Combinational_Single:
    Process(LCD_Read,
            LCD_Write,
            LCD_Select,
            CycleCounter,
            ADR_I,
            State)
    Begin
        ------------------------------------------------------------------------
        -- Default outputs
        LCD_RW             <= LCD_Read;     -- Read if 1 write if zero
        LCD_RS             <= ADR_I;
        ACK_O              <= '0';
        State_Next         <= State;
        CycleCounter_Reset <= '0';
        LCD_E              <= '0';
        ------------------------------------------------------------------------

        ------------------------------------------------------------------------
        Case State Is
           When State_AddressSetup =>
                CycleCounter_Reset <= '1';
                If LCD_Select = '1' Then                --Address is being setup in this phase
                   State_Next <= State_SetupDelay1;
                End If;

           When State_SetupDelay1 =>
                State_Next <= State_SetupDelay2;

           When State_SetupDelay2 =>
                State_Next <= State_Enable_LCD;

           When State_Enable_LCD =>
                LCD_E    <= '1';
                If CycleCounter(3 Downto 0) = "1111" Then
                   State_Next <= State_ACK;
                End If;

           When State_ACK =>                          -- Assert ACK_O, Master will read the data one clock cycle later
                LCD_E      <= LCD_Read;               -- For a write, pull down the Enable one cycle before address, Data change.
                ACK_O      <= '1';                    -- For a read, hold the enable until the read is complete.
                CycleCounter_Reset <= '1';
                State_Next <= State_PostACKDelay;

           When State_PostACKDelay =>
                If CycleCounter = "1010000" Then
                   State_Next <= State_AddressSetup;
                End If;

           When Others =>
                State_Next <= State_AddressSetup;

        End Case;
    End Process;
    ----------------------------------------------------------------------------

    ----------------------------------------------------------------------------
    FSM_Synchronous:
    Process(CLK_I)
    Begin
        If Rising_Edge(CLK_I) Then
           If RST_I = '1' Then
              State  <= State_AddressSetup;
           Else
              State <= State_Next;
           End If;
        End If;
    End Process;
    ----------------------------------------------------------------------------

    ----------------------------------------------------------------------------
    CycleCounterControl:
    Process(CLK_I)
    Begin
        If Rising_Edge(CLK_I) Then
           If (RST_I = '1') Or (CycleCounter_Reset = '1') Then
              CycleCounter <= (Others => '0');
           Else
              CycleCounter <= CycleCounter + 1;
           End If;
        End If;
    End Process;
    ----------------------------------------------------------------------------
End RTL;

--------------------------------------------------------------------------------
--  Note: max clock freq 50MHz
--------------------------------------------------------------------------------
--  Read Cycle:                    Write Cycle:                    State
--------------------------------------------------------------------------------
--  RS   Set                       RS   Set                   State_AddressSetup
--  RW   High                      RW   Low
--  Data Out Setup                 Data Out HiZ
--  LCD_E Low
--------------------------------------------------------------------------------
--  LCD_E Low   ACK LOW     ( 1 Cycle )
--------------------------------------------------------------------------------
--  LCD_E High  ACK LOW     (12 Cycles)
--------------------------------------------------------------------------------
--  LCD_E High  ACK HIGH    ( 1 Cycle )
--------------------------------------------------------------------------------
--  LCD_E Low   ACK LOW     ( 1 Cycle )   MCU Captures Data
--------------------------------------------------------------------------------
--  Done - Total 15 Clock Cycles (300 ns)
--
----------------------------------
--  Cannot start another cycle for another 1600ns - another 80 cycles or so




