--------------------------------------------------------------------------------
Library IEEE;
Use IEEE.Std_Logic_1164.all;
Use IEEE.Std_Logic_Unsigned.All;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
Entity WB_LCDCTRL_SRAM Is Port
   (
     -- Wishbone Interface for LCD controller
     STB_I           : IN    STD_LOGIC;
     CYC_I           : IN    STD_LOGIC;
     ACK_O           : OUT   STD_LOGIC;
     ADR_I           : IN    STD_LOGIC;
     WE_I            : IN    STD_LOGIC;
     DAT_I           : IN    STD_LOGIC_VECTOR(7 DOWNTO 0);
     DAT_O           : OUT   STD_LOGIC_VECTOR(7 DOWNTO 0);
     CLK_I           : IN    STD_LOGIC;
     RST_I           : IN    STD_LOGIC;

     -- INTERFACE TO CTRL/LCD
     LCD_E           : OUT   STD_LOGIC;
     MEM_CS          : OUT   STD_LOGIC;
     MEM_ADDR        : OUT   STD_LOGIC_VECTOR(16 DOWNTO 0);
     MEM0_DATA       : INOUT STD_LOGIC_VECTOR(7 DOWNTO 0);
     MEM1_DATA       : INOUT STD_LOGIC_VECTOR(7 DOWNTO 0);
     MEM0_OE         : OUT   STD_LOGIC;
     MEM1_OE         : OUT   STD_LOGIC;
     MEM0_WE         : OUT   STD_LOGIC;
     MEM1_WE         : OUT   STD_LOGIC;

     -- INTERFACE TO CTRL CONTROLLER
     CTRL_CE         : IN    STD_LOGIC;
     CTRL_A          : IN    STD_LOGIC_VECTOR(16 DOWNTO 0);
     CTRL0_D         : INOUT STD_LOGIC_VECTOR(7 DOWNTO 0);
     CTRL1_D         : INOUT STD_LOGIC_VECTOR(7 DOWNTO 0);
     CTRL0_OE        : IN    STD_LOGIC;
     CTRL1_OE        : IN    STD_LOGIC;
     CTRL0_WE        : IN    STD_LOGIC;
     CTRL1_WE        : IN    STD_LOGIC
   );
End WB_LCDCTRL_SRAM;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
Architecture RTL Of WB_LCDCTRL_SRAM Is
    Type TState Is
        (
           State_AddressSetup,
           State_SetupDelay1,
           State_SetupDelay2,
           State_Enable_LCD,
           State_ACK,
           State_PostACKDelay
        );

    Signal   State                : TState;
    Signal   State_Next           : TState;
    Signal   LCD_Read             : Std_Logic;
    Signal   LCD_Write            : Std_Logic;
    Signal   LCD_Select           : Std_Logic;
    Signal   LCD_RW               : Std_Logic;   -- 0:Write 1:Read
    Signal   LCD_RS               : Std_Logic;   -- 0:Control Register  1:Data Register
    Signal   CycleCounter         : Std_Logic_Vector(6 DownTo 0);
    Signal   CycleCounter_Reset   : Std_Logic;

Begin
    ----------------------------------------------------------------------------
    -- Permanent Assignments
    -- Driven outputs
    Mem_ADDR(16 DownTo 2) <= CTRL_A(16 DownTo 2);

    Mem0_OE <= CTRL0_OE;
    Mem1_OE <= CTRL1_OE;
    Mem0_WE <= CTRL0_WE;
    Mem1_WE <= CTRL1_WE;
    Mem_CS  <= CTRL_CE Or LCD_Select; -- disable memory if lcd_select is true

    DAT_O   <= Mem0_DATA;
    ----------------------------------------------------------------------------

    ----------------------------------------------------------------------------
    LCD_Select <= STB_I And CYC_I;
    LCD_Read   <= STB_I And CYC_I And (Not WE_I);
    LCD_Write  <= STB_I And CYC_I And (    WE_I);
    ----------------------------------------------------------------------------

    ----------------------------------------------------------------------------
    Multiplexers:
    Process (Mem0_DATA,
             Mem1_DATA,
             CTRL0_D,
             CTRL1_D,
             CTRL_CE,
             CTRL_A,
             DAT_I,
             CTRL0_OE,
             CTRL1_OE,
             LCD_RS,
             LCD_RW,
             LCD_Select,
             LCD_Write)
    Begin
        If CTRL1_OE = '0' Then --Reading from memory
           Mem1_DATA <= (Others => 'Z');
           CTRL1_D   <= Mem1_DATA;
        Else                   --Writing to memory
           Mem1_DATA <= CTRL1_D;
           CTRL1_D   <= (Others => 'Z');
        End If;

        If LCD_Select = '0' Then -- Leave it to the CTRL controller.
           Mem_ADDR(1 DownTo 0) <= CTRL_A(1 DownTo 0);

           If CTRL0_OE = '0' Then --Reading from memory
              Mem0_DATA <= (Others => 'Z');
              CTRL0_D   <= Mem0_DATA;
           Else                   --Writing to memory
              Mem0_DATA <= CTRL0_D;
              CTRL0_D   <= (Others => 'Z');
           End If;
        Else
           Mem_ADDR(0) <= LCD_RS;
           Mem_ADDR(1) <= LCD_RW;
           CTRL0_D     <= (Others => 'Z');

           If LCD_Write = '1' Then
              Mem0_Data <= DAT_I;
           Else
              Mem0_Data <= (Others => 'Z');
           End If;
        End If;
    End Process;
    ----------------------------------------------------------------------------

    ----------------------------------------------------------------------------
    FSM_Combinational_Single:
    Process(LCD_Read,
            LCD_Write,
            LCD_Select,
            CycleCounter,
            ADR_I,
            State)
    Begin
        ------------------------------------------------------------------------
        -- Default outputs
        LCD_RW             <= LCD_Read;     -- Read if 1 write if zero
        LCD_RS             <= ADR_I;
        ACK_O              <= '0';
        State_Next         <= State;
        CycleCounter_Reset <= '0';
        LCD_E              <= '0';
        ------------------------------------------------------------------------

        ------------------------------------------------------------------------
        Case State Is
           When State_AddressSetup =>
                CycleCounter_Reset <= '1';
                If LCD_Select = '1' Then                --Address is being setup in this phase
                   State_Next <= State_SetupDelay1;
                End If;

           When State_SetupDelay1 =>
                State_Next <= State_SetupDelay2;

           When State_SetupDelay2 =>
                State_Next <= State_Enable_LCD;

           When State_Enable_LCD =>
                LCD_E    <= '1';
                If CycleCounter(3 Downto 0) = "1111" Then
                   State_Next <= State_ACK;
                End If;

           When State_ACK =>                          -- Assert ACK_O, Master will read the data one clock cycle later
                LCD_E      <= LCD_Read;               -- For a write, pull down the Enable one cycle before address, Data change.
                ACK_O      <= '1';                    -- For a read, hold the enable until the read is complete.
                CycleCounter_Reset <= '1';
                State_Next <= State_PostACKDelay;

           When State_PostACKDelay =>
                If CycleCounter = "1010000" Then
                   State_Next <= State_AddressSetup;
                End If;

           When Others =>
                State_Next <= State_AddressSetup;

        End Case;
    End Process;
    ----------------------------------------------------------------------------

    ----------------------------------------------------------------------------
    FSM_Synchronous:
    Process(CLK_I,RST_I,State_Next)
    Begin
        If Rising_Edge(CLK_I) Then
           If RST_I = '1' Then
              State  <= State_AddressSetup;
           Else
              State <= State_Next;
           End If;
        End If;
    End Process;
    ----------------------------------------------------------------------------

    ----------------------------------------------------------------------------
    CycleCounterControl:
    Process(CLK_I,RST_I,CycleCounter_Reset)
    Begin
        If Rising_Edge(CLK_I) Then
           If (RST_I = '1') Or (CycleCounter_Reset = '1') Then
              CycleCounter <= (Others => '0');
           Else
              CycleCounter <= CycleCounter + 1;
           End If;
        End If;
    End Process;
    ----------------------------------------------------------------------------
End RTL;

--------------------------------------------------------------------------------
--  Note: max clock freq 50MHz
--------------------------------------------------------------------------------
--  Read Cycle:                    Write Cycle:                    State
--------------------------------------------------------------------------------
--  RS   Set                       RS   Set                   State_AddressSetup
--  RW   High                      RW   Low
--  Data Out Setup                 Data Out HiZ
--  LCD_E Low
--------------------------------------------------------------------------------
--  LCD_E Low   ACK LOW     ( 1 Cycle )
--------------------------------------------------------------------------------
--  LCD_E High  ACK LOW     (12 Cycles)
--------------------------------------------------------------------------------
--  LCD_E High  ACK HIGH    ( 1 Cycle )
--------------------------------------------------------------------------------
--  LCD_E Low   ACK LOW     ( 1 Cycle )   MCU Captures Data
--------------------------------------------------------------------------------
--  Done - Total 15 Clock Cycles (300 ns)
--
----------------------------------
--  Cannot start another cycle for another 1600ns - another 80 cycles or so




